from collections.abc import Hashable, Iterable, Mapping, Sequence
from copy import deepcopy
from functools import reduce
from typing import Any, TypeVar, cast, overload

from .display import identity

T = TypeVar("T")
K = TypeVar("K", bound=Hashable)
V = TypeVar("V")

JsonValue = Hashable | Sequence["JsonValue"] | Mapping[Hashable, "JsonValue"]
DictType = Mapping[Hashable, "JsonValue"]


def _dedup[T](x: Iterable[T]) -> list[T]:
    tracker: set[T] = set()
    result: list[T] = []
    for elem in x:
        if elem in tracker:
            continue
        tracker.add(elem)
        result.append(elem)
    return result


@overload
def dedup[T](x: list[T]) -> list[T]: ...
@overload
def dedup[T](x: tuple[T, ...]) -> tuple[T, ...]: ...
@overload
def dedup[T](x: set[T]) -> set[T]: ...
def dedup(x):  # type: ignore
    if isinstance(x, list):
        return _dedup(x)
    if isinstance(x, tuple):
        return tuple(_dedup(x))
    if isinstance(x, set):
        return set(x)
    raise TypeError


def make_hashable(obj: Any) -> Hashable:
    """
    Convert any object to a hashable representation.

    Improved: Recursively converts nested structures for true hashability.
    """

    if isinstance(obj, (list, tuple, set)):
        return tuple(make_hashable(item) for item in obj)

    if isinstance(obj, Hashable):
        return obj

    if isinstance(obj, dict):
        return tuple(((make_hashable(k), make_hashable(v)) for k, v in obj.items()))

    return str(obj)


def _should_include(item: Any, existing: Iterable, add_duplicates: bool) -> bool:
    """Determine if an item should be included when merging sequences."""
    if add_duplicates:
        return True

    if isinstance(existing, (list, tuple, set)):
        return item not in existing
    return item != existing


def _merge_sequence_with_value(
    seq: Sequence[T] | set[T], value: T, add_duplicates: bool
) -> Sequence[T] | set[T]:
    """Merge a sequence with a single value."""
    if isinstance(seq, list):
        return seq if not _should_include(value, seq, add_duplicates) else [*seq, value]

    if isinstance(seq, tuple):
        return seq if not _should_include(value, seq, add_duplicates) else (*seq, value)

    if isinstance(seq, set):
        hashable_val = make_hashable(value)
        return (  # type: ignore
            seq
            if not _should_include(hashable_val, seq, add_duplicates)
            else seq | {cast(T, hashable_val)}
        )

    return seq


def _merge_value_with_sequence(
    value: T, seq: Sequence[T] | set[T], add_duplicates: bool
) -> Sequence[T] | set[T]:
    """Merge a sequence with a single value."""
    if isinstance(seq, list):
        return seq if not _should_include(value, seq, add_duplicates) else [value, *seq]

    if isinstance(seq, tuple):
        return seq if not _should_include(value, seq, add_duplicates) else (value, *seq)

    if isinstance(seq, set):
        hashable_val = cast(T, make_hashable(value))
        return seq | {hashable_val}

    return seq


def _merge_two_sequences(
    left: Sequence[T] | set[T], right: Sequence[T] | set[T], add_duplicates: bool
) -> Sequence[T] | set[T]:
    """Merge two sequences/collections."""
    if isinstance(left, set):
        right_hashable = (make_hashable(item) for item in right)
        return {*left, *right_hashable}  # type: ignore

    _maybe_dedup = identity if add_duplicates else dedup
    return left.__class__(_maybe_dedup(list(left) + list(right)))  # type: ignore


def join_as_sequence(
    left: JsonValue,
    right: JsonValue,
    default_type: type[list] | type[tuple] | type[set] = tuple,
    add_duplicates: bool = True,
) -> Sequence | set:
    """
    Join two values into a sequence.

    Improved: More consistent type handling and clearer logic flow.
    """
    if isinstance(left, (list, tuple, set)) and isinstance(right, (list, tuple, set)):
        return _merge_two_sequences(left, right, add_duplicates)

    if isinstance(left, (list, tuple, set)):
        return _merge_sequence_with_value(left, right, add_duplicates)

    if isinstance(right, (list, tuple, set)):
        result = _merge_value_with_sequence(left, right, add_duplicates)
        return result

    if add_duplicates or left != right:
        return default_type((left, right))
    return default_type((left,))


def merge_dicts(  # noqa: C901 (complexity)
    *dicts: DictType,
    default_sequence_type: type[list] | type[tuple] | type[set] = tuple,
    add_duplicates: bool = True,
) -> dict[Hashable, Any]:
    """
    Deep merge multiple dictionaries.
    """

    def merge_values(left_val: Any, right_val: Any) -> Any:  # noqa: PLR0911 (too many return statements)
        """Merge two values, handling nested structures."""
        if left_val == right_val:
            return left_val.copy() if isinstance(left_val, dict) else left_val

        if isinstance(left_val, dict) and isinstance(right_val, dict):
            return merge_dicts(
                left_val,
                right_val,
                default_sequence_type=default_sequence_type,
                add_duplicates=add_duplicates,
            )

        if left_val is None:
            return right_val
        if right_val is None:
            return left_val

        if isinstance(left_val, dict):
            return left_val.copy()
        if isinstance(right_val, dict):
            return right_val.copy()

        return join_as_sequence(
            left_val, right_val, default_type=default_sequence_type, add_duplicates=add_duplicates
        )

    def merge_two(left: DictType, right: DictType) -> dict[Hashable, Any]:
        """Merge two dictionaries."""
        result = dict(left)

        for key, right_value in right.items():
            if key not in result:
                result[key] = right_value
            else:
                result[key] = merge_values(result[key], right_value)

        return result

    if not dicts:
        return {}
    if len(dicts) == 1:
        return dict(dicts[0]).copy()

    return cast(dict[Hashable, Any], reduce(merge_two, dicts))


def deep_get(dictionary: DictType, *keys: Hashable, default: Any = None) -> Any:
    """Safely get a value from a nested dictionary."""
    current = dictionary
    for key in keys:
        if not isinstance(current, dict) or key not in current:
            return default
        current = current[key]
    return current


def deep_set(dictionary: DictType, value: Any, *keys: Hashable) -> DictType:
    """Set a value in a nested dictionary (returns new dict)."""

    result = dict(deepcopy(dictionary))
    print(id(dictionary), id(result))
    if not keys:
        return result
    current: dict = result

    for key in keys[:-1]:
        if key not in current or not isinstance(current[key], dict):
            current[key] = {}
        current = current[key]

        # if key not in current or not isinstance(value := current[key], dict):  # type: ignore
        #     current[key] = {}  # type: ignore
        # current = current[key]  # type: ignore

    # current[keys[-1]] = value  # type: ignore
    # return result.copy()  # type: ignore

    current[keys[-1]] = value
    return result



import os
from datetime import datetime
from pathlib import Path

from .io import ensure_path


def time_created(path: Path | str) -> float:
    return ensure_path(path).stat().st_ctime


def format_time(raw_time: float, places: int = 3) -> str:
    dt = datetime.fromtimestamp(raw_time)
    if places > 0:
        idx: int | None = min(0, int(places) - 6) or None
        return dt.strftime("%Y-%m-%d_%H:%M:%S.%f")[:idx]
    return dt.strftime("%Y-%m-%d_%H:%M:%S")


def time_created_readable(path: Path | str, places: int = 3) -> str:
    # time_created: time.struct_time = time.strptime(time.ctime(time_created(path)))
    return format_time(time_created(path), places=places)


def time_modified(path: Path | str) -> float:
    return ensure_path(path).stat().st_mtime


def time_modified_readable(path: Path | str, places: int = 3) -> str:
    # time_modified: time.struct_time = time.strptime(time.ctime(time_modified(path)))
    return format_time(time_modified(path), places=places)


def first_newer(file1: str | Path, file2: str | Path | tuple[Path, ...] | tuple[str, ...]) -> bool:
    m1 = os.path.getmtime(file1)

    if isinstance(file2, tuple):
        m2 = max(os.path.getmtime(f) for f in file2)
    else:
        m2 = os.path.getmtime(file2)
    return m1 > m2



from abc import ABC
from collections import defaultdict
from collections.abc import Callable
from typing import Self, TypeVar, cast, get_args

__all__ = ("FrozenDefaultDict", "FrozenDict", "Immutable", "is_immutable")

T = TypeVar("T")
K = TypeVar("K")


type Immutable = int | float | tuple | str | FrozenDict | FrozenDefaultDict
_IMMUTABLE_TYPES = get_args(Immutable)


def is_immutable(ob: object) -> bool:
    return isinstance(ob, _IMMUTABLE_TYPES)


def freeze(ob: object) -> object:
    if is_immutable(ob):
        return ob
    if isinstance(ob, list | set):
        return tuple(ob)
    if isinstance(ob, dict):
        return FrozenDict(ob)
    if isinstance(ob, defaultdict):
        return FrozenDefaultDict.from_defaultdict(ob)
    raise NotImplementedError
    # TODO: overload to convert any object into its immutable counterpart (if not already)


class _FrozenDictBase(ABC):
    @property
    def _immutable_error(self) -> TypeError:
        return TypeError(f"{self.__class__.__name__} is immutable.")

    def update(self, _) -> None:
        raise self._immutable_error

    def setdefault(self, _, __) -> None:
        raise self._immutable_error

    def pop(self) -> None:
        raise self._immutable_error

    def popitem(self, _) -> None:
        raise self._immutable_error

    def __setitem__(self, _, __) -> None:
        raise self._immutable_error

    def __delitem__(self, _) -> None:
        raise self._immutable_error


class FrozenDict(_FrozenDictBase, dict[K, T]):
    ...  # type: ignore
    # see https://pypi.org/project/frozendict/, but make stricter


class FrozenDefaultDict[K, T](_FrozenDictBase, defaultdict[K, T]):  # type: ignore
    def __init__(self, default_factory: Callable[[], T], dictionary: dict[K, T]):
        super().__init__(default_factory, dictionary)

    @classmethod
    def from_defaultdict(cls, dd: defaultdict[K, T]) -> Self:
        _factory = cast(Callable[[], T], dd.default_factory)
        return cls(_factory, dict(dd.items()))

    def __repr__(self) -> str:
        default = self.default_factory
        result = default.__name__ if callable(default) else repr(default)
        return f"{self.__class__.__name__}({dict(self)}, default={result})"

    def __getitem__(self, key: K) -> T:
        if key in self:
            return super().__getitem__(key)
        return cast(Callable, self.default_factory)()



INF = float("inf")


def round_to(
    num: int | float,
    *,
    modulus: int | float,
    min_val: int | float = -INF,
    max_val: int | float = INF,
) -> int:
    """For numbers greater than min_val, round to the nearest multiple of 5."""
    if num < min_val:
        return round(num)
    q, r = divmod(num, modulus)
    return int(modulus * (q + int(r > (modulus / 2))))


def round5(num: int | float, min_val: int | float = 15) -> int:
    """For numbers greater than min_val, round to the nearest multiple of 5."""
    if num < min_val:
        return round(num)
    q, r = divmod(num, 5)
    return int(5 * (q + int(r > 2)))


def evenly_spaced(
    start: float,
    end: float,
    steps: int,
    reversed: bool = False,
) -> list[float]:
    if start > end:
        start, end = end, start
        reversed = True
    start += 1e-7
    end -= 1e-7
    step_size = (end - start) / steps
    numbers = [start + step_size * i for i in range(steps)]
    if reversed:
        numbers.reverse()
    return numbers


def ihash(i: int) -> str:
    return str(hash(str(i)))[-4:]



import json
import os
import re
import shutil
from pathlib import Path
from typing import Literal, cast

from .fp import tmap
from .timestamps import make_timestamp
from .types import JsonContainer
from .typing import JSONDict, JSONList


def ensure_path(p: Path | str) -> Path:
    if isinstance(p, Path):
        return p
    return Path(p)


def list_full(directory: str | Path, ending: str = "") -> list[Path]:
    directory = ensure_path(directory)
    return sorted([directory / file for file in os.listdir(directory) if file.endswith(ending)])


def read_raw(json_path: Path) -> str:
    with open(json_path, encoding="utf-8") as f:
        return f.read()


def back_up_json(original_path: Path, backup_dir: Path | None) -> None:
    if not backup_dir:
        return
    backup_path = backup_dir / original_path.name.replace(
        ".json", f"__{make_timestamp(places=4)}.json"
    )
    if original_path.exists():
        shutil.copy(original_path, backup_path)


def read_json(json_path: Path, backup_dir: Path | None = None) -> JSONDict | JSONList:
    back_up_json(json_path, backup_dir)
    with open(json_path, encoding="utf-8") as f:
        return json.load(f)


def write_json(
    python_obj: JSONDict | JSONList | str | bytes,
    json_path: Path,
    backup_dir: Path | None = None,
    mode: Literal["w", "a"] = "w",
) -> None:
    if mode not in {"w", "a"}:
        raise ValueError(f"Invalid mode: '{mode}'")
    back_up_json(json_path, backup_dir)
    if mode == "a":
        if not json_path.exists():
            existing: dict | list | None = None
        else:
            existing = read_json(json_path)
        if isinstance(existing, list | None) and isinstance(python_obj, list):
            python_obj = (existing or []) + python_obj
        elif isinstance(existing, dict | None) and isinstance(python_obj, dict):
            python_obj = (existing or {}) | python_obj
        else:
            raise ValueError("Data types do not match.")
    if isinstance(python_obj, bytes):
        python_obj = python_obj.decode()
    if isinstance(python_obj, str):
        python_obj = json.loads(python_obj)
    with open(json_path, "w", encoding="utf-8") as f:
        json.dump(python_obj, f, ensure_ascii=False, indent=4)


def write_raw(text: str, file_path: Path) -> None:
    with open(file_path, "w", encoding="utf-8") as f:
        f.write(text)


def write_raw_bytes(text: bytes, json_path: Path) -> None:
    with open(json_path, "wb") as f:
        f.write(text)


def list_names(directory: Path, ending: str) -> tuple[str, ...]:
    return tmap(
        lambda p: p.name.replace(ending, ""),
        list_full(directory, ending=ending),
    )


def name_from_full(p: Path | str, ending: str = ".json") -> str:
    if isinstance(p, Path):
        return p.name.replace(ending, "")
    return p.split("/")[-1].replace(ending, "")


def remove_all_under(dir_: Path) -> None:
    for item in dir_.iterdir():
        if item.is_file() or item.is_symlink():
            item.unlink()
        elif item.is_dir():
            shutil.rmtree(item)


def read_jsonc(path: Path) -> JsonContainer:
    def _strip_comments(s: str) -> str:
        s = s.strip()
        if s.startswith("//"):
            return ""
        return search.group(1) if (search := re.search(r"^(.+?[,\"\}\]\d e]) *//.*$", s)) else s

    return cast(
        JsonContainer,
        json.loads("".join(map(_strip_comments, path.read_text().splitlines()))),
    )



from abc import ABC, abstractmethod
from collections.abc import Iterable
import glob
from pathlib import Path
import re
from typing import Self

import git
from git.remote import PushInfoList
from loguru import logger
from pydantic import DirectoryPath, Field

from .fp import lfilter, lmap, negate

from .exceptions import GitStatusError
from .pydantic import BaseModelRW
from .timestamps import make_timestamp
from .typing import ensure_tuple

# class PathsManager:
#     @abstractmethod
#     def setup(self) -> None: ...

#     @classmethod
#     @abstractmethod
#     def auto(cls, root_dir: Path): ...

#     @classmethod
#     @abstractmethod
#     def read(cls, config_file_path: Path) -> Self: ...

#     @abstractmethod
#     def write(self, config_file_path: Path) -> None: ...


def glob_extension(extension: str = "*", path: Path | None = None, ignore: re.Pattern[str] | str | None = None) -> list[Path]:
    """Search a directory recursively for all files ending with 'extension' and not matching 'ignore'.
    
        Defaults to the current directory.
    """
    if not path:
        path = Path.cwd()
    expression = f"**/*.{extension}"
    result: list[str] = glob.glob(expression, recursive=True)
    if ignore:
        _ignore = re.compile(ignore) if isinstance(ignore, str) else ignore
        result = lfilter(negate(_ignore.search), result)
    return lmap(Path, result)


class PathsManager(BaseModelRW, ABC):
    git_root: DirectoryPath | None = Field(default=None)
    # @property
    # @abstractmethod
    # def git_root(self) -> Path | None:
    #     raise NotImplementedError

    @property
    def repo(self) -> git.Repo | None:
        if not self.git_root:
            return None
        dot_git = self.git_root / ".git"
        if not dot_git.exists():
            return git.Repo.init(self.git_root)
        return git.Repo(self.git_root, search_parent_directories=True)

    @classmethod
    @abstractmethod
    def create(cls, *args, **kwargs) -> Self:
        """
        Create directories first if they do not already exist.
        """
        ...

    @classmethod
    @abstractmethod
    def auto(cls, *args, **kwargs) -> Self:
        """
        Create directories and files in the default location if they do not already exist.
        """
        ...

    @classmethod
    @abstractmethod
    def read(cls, config_file_path: Path) -> Self: ...

    @abstractmethod
    def write(self, config_file_path: Path) -> None: ...

    @staticmethod
    def ensure_file_exists(
        p: Path, content_if_empty: str = "", root_must_exist: bool = True
    ) -> Path:
        PathsManager.ensure_parents_exist(p, root_must_exist)
        if not p.exists():
            p.touch()
            if content_if_empty:
                p.write_text(content_if_empty)
        return p

    @staticmethod
    def ensure_directory_exists(p: Path, root_must_exist: bool = True) -> Path:
        PathsManager.ensure_parents_exist(p, root_must_exist)
        if not p.exists():
            p.mkdir()
        return p

    @staticmethod
    def ensure_parents_exist(p: Path, root_must_exist: bool = True) -> None:
        if not (root := Path(p.root)).exists():
            if root_must_exist:
                raise OSError(f"Root path does not exist: {root}")
            root.mkdir()
        for parent in reversed(p.parents):
            if not parent.exists():
                parent.mkdir()

    def assert_clean(self) -> None:
        if not self.is_clean:
            raise GitStatusError(f"Git repo at {self.git_root} must be clean.")

    @property
    def is_clean(self) -> bool:
        if not self.repo:
            return False
        return not self.repo.is_dirty()

    def commit_data(
        self,
        path_or_paths: Path | Iterable[Path] = tuple(),
        *,
        message: str | None = None,
        push: bool = True,
        remote_name: str = "origin",
        remote_url: str | None = None,
    ) -> tuple[git.Commit | None, PushInfoList | None]:
        def make_default_message() -> str:
            return f"automatic commit by '{self.__class__.__name__}' at {make_timestamp()}"

        repo = self.repo
        if not repo:
            return None, None

        for path in ensure_tuple(path_or_paths or self.git_root):
            repo.git.add(str(path))
        commit = repo.index.commit(message or make_default_message())

        push_info: PushInfoList | None = None
        if push:
            try:
                if remote_url:
                    remote = repo.create_remote(name=remote_name, url=remote_url)
                else:
                    remote = repo.remote(name=remote_name)
                push_info = remote.push()
            except Exception as e:
                logger.info(f"Push error: {e}")

        return commit, push_info



import json
import re
from collections.abc import Callable, Iterable
from typing import Literal, Protocol, Self, TypeAlias, cast

from .fp import lmap
from .typing import JSONDict


class DateProtocol(Protocol):
    year: int
    month: int
    day: int

    @classmethod
    def parse(cls, date_string: str) -> Self: ...

    def __str__(self) -> str: ...


class TimeProtocol(Protocol):
    hour: int
    minute: int
    second: float

    @classmethod
    def parse(cls, time_string: str) -> Self: ...

    def __str__(self) -> str: ...


MixedValidated: TypeAlias = (
    str
    | bool
    | int
    | float
    | TimeProtocol
    | DateProtocol
    | tuple[str, ...]
    | tuple[str, str]
    | tuple[float, float]
    | None
)
PromptTypeName = Literal[
    "string",
    "boolean",
    "integer",
    "float",
    "minutes",
    "time",
    "positiveScore",
    "negativeScore",
    "date",
    "stringtuple",
]


def indent_lines(lines: Iterable[object], indent_size: int = 4) -> str:
    joiner = "\n" + indent_size * " "
    return joiner + joiner.join(map(str, lines))


def flexsplit(s: str) -> list[str]:
    return re.split(r"[ ,]+", s)


def split_sequence_string(seq: str) -> tuple[str, str, int]:
    _start: str | None
    _end: str | None
    _step: str | None
    match seq.count(":"):
        case 2:
            _start, _end, _step = seq.split(":")
        case 1:
            _start, _end = seq.split(":")
            _step = None
        case 0:
            _start, _end, _step = seq, seq, None
        case _:
            raise ValueError
    return (_start or "").upper(), (_end or "").upper(), int(_step or 1)


def parse_sequence(s: str) -> list[str]:
    if not s:
        return []

    def interpolate(_s: str) -> list[str]:
        start, end, step = split_sequence_string(_s)
        if end.isnumeric():
            return lmap(str, range(int(start or 1), int(end) + 1, step))
        elif end.isalpha():
            start = start or "A"
            assert len(start) == len(end) == 1
            return lmap(chr, range(ord(start), ord(end) + 1, step))
        else:
            pattern = re.compile(r"[A-Z]\d+$")
            letter = start[0]
            assert re.match(pattern, start) and re.match(pattern, end) and (letter == end[0])
            start, end = start[1:], end[1:]
            return [f"{letter}{i}" for i in range(int(start), int(end) + 1, step)]

    segments: list[str] = []
    for subseq in re.split(" *, *", s.strip()):
        segments.extend(interpolate(subseq))
    return segments


def cast_to_bool(s: str | bool) -> bool:
    if isinstance(s, bool):
        return s
    s = s.strip().lower()
    if s.startswith(("y", "t")):
        return True
    if s.startswith(("f", "n")):
        return False
    raise ValueError(f"Ambiguous input for 'cast_to_bool': '{s}'")


def cast_to_int(s: str | bool) -> int:
    if isinstance(s, bool):
        return int(s)
    return int(s.strip())


def cast_to_float(s: str | bool) -> float:
    if isinstance(s, bool):
        return float(s)
    return float(s.strip())


def cast_to_minutes(s: str | bool) -> int:
    if isinstance(s, bool):
        raise TypeError
    s = s.strip()
    if ":" in s:
        if s.count(":") > 1:
            raise ValueError
        hours, minutes = s.split(":")
        return 60 * int(hours) + int(minutes)
    return int(s)


def cast_to_positive_score(s: str | bool) -> float:
    if isinstance(s, bool):
        raise TypeError
    score = float(s.strip())
    if not 0.0 <= score <= 5.0:
        raise ValueError
    return score


def cast_to_negative_score(s: str | bool) -> float:
    if isinstance(s, bool):
        raise TypeError
    score = float(s.strip())
    if not -5.0 <= score <= 0.0:
        raise ValueError
    return score


def cast_to_date(s: str | bool, date_class: type[DateProtocol]) -> DateProtocol:
    if isinstance(s, bool):
        raise TypeError
    return date_class.parse(s)


def cast_to_stringtuple(s: str | bool) -> tuple[str, ...]:
    if isinstance(s, bool):
        raise TypeError
    if not s:
        return tuple()
    return cast(tuple[str, ...], tuple(re.split(r"[ ,;]+", s)))


def cast_as(
    input_type: PromptTypeName,
    null_strings: Iterable[str] | None = ("none", "null", "skip", "pass"),
    optional: bool = True,
) -> Callable[[str | bool], MixedValidated]:
    dispatch: dict[
        PromptTypeName,
        Callable[[str | bool], MixedValidated],
    ] = {
        "boolean": cast_to_bool,
        "integer": cast_to_int,
        "float": cast_to_float,
        "minutes": cast_to_minutes,
        # "time": Time.model_validate, TODO: add later via injection
        "positiveScore": cast_to_positive_score,
        "negativeScore": cast_to_negative_score,
        # "date": cast_to_date, TODO: add later via injection
        "stringtuple": cast_to_stringtuple,
    }
    caster = dispatch[input_type]

    null_strings = set(null_strings or set())

    def nullable_caster(s: str | bool) -> MixedValidated:
        if str(s).lower() in null_strings:
            return None
        try:
            return caster(s)
        except ValueError as e:
            if optional:
                return None
            raise e

    return nullable_caster


def as_json(d: JSONDict) -> str:
    return json.dumps(d, ensure_ascii=False, indent=4)


def transfer_case(reference: str, candidate: str) -> str:
    if reference.istitle():
        return candidate.title()
    if reference.islower():
        return candidate.lower()
    if reference.isupper():
        return candidate.upper()
    return candidate



"""
Idea: make BaseSequence[C, T] type,
    where C is list|set|tuple and T is the element type

Idea: Make a BaseTuple class -> should accept a TypeVarTuple,
    for a flexible number of type arguments
"""

import json
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterable, Mapping
from pathlib import Path
from types import GenericAlias
from typing import (
    Any,
    Literal,
    Protocol,
    Self,
    TypeVar,
    cast,
    get_args,
    runtime_checkable,
)

from pydantic import (
    BaseModel,
    ConfigDict,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    TypeAdapter,
)
from pydantic.config import ExtraValues
from pydantic.fields import FieldInfo
from pydantic.json_schema import (
    DEFAULT_REF_TEMPLATE,
    GenerateJsonSchema,
    JsonSchemaMode,
    JsonSchemaValue,
)
from pydantic.main import IncEx
from pydantic_core import CoreSchema, core_schema

from .io import JSONDict, read_json, write_json

T = TypeVar("T")
K_ = TypeVar("K_", bound=Hashable)
V_ = TypeVar("V_")
Mode = Literal["json", "python"] | str


VALID_MODES: set[Literal["json", "python"]] = {"json", "python"}


@runtime_checkable
class ComplexValidatedProtocol(Protocol):
    """
    Protocol defining minimum shared behavior between AbstractCustom, its children, and BaseModelRW.
    """

    @classmethod
    def read_json_file(cls, source: Path) -> Self: ...

    def write_json_file(self, destination: Path) -> Path: ...


class BaseModelRW(BaseModel):
    @classmethod
    def read_json_file(cls, source: Path) -> Self:
        return cls.model_validate(read_json(source))

    def write_json_file(self, destination: Path, by_alias: bool = True) -> Path:
        write_json(self.model_dump(mode="json", by_alias=by_alias), destination)
        return destination


class AbstractCustom(ABC):
    model_config: ConfigDict = ConfigDict()
    __pydantic_core_schema__: CoreSchema

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.model_validate,
            core_schema.dict_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        return json_schema

    def __init__(self, *args, **kwargs) -> None:
        raise TypeError(f"Use `model_validate` to instantiate {self.__class__.__name__}")

    @classmethod
    @abstractmethod
    def get_adapter(cls) -> TypeAdapter: ...

    @abstractmethod
    def __repr__(self) -> str: ...

    @abstractmethod
    def __str__(self) -> str: ...

    @classmethod
    def model_fields(cls) -> dict[str, FieldInfo]:
        return {}

    @property
    def model_extra(self) -> dict[str, Any] | None:
        return None

    @property
    def model_fields_set(self) -> set[str]:
        return set()

    @classmethod
    def model_construct(cls, _fields_set: set[str] | None = None, **values: Any) -> Self:
        raise NotImplementedError

    def model_copy(self, *, update: Mapping[str, Any] | None = None, deep: bool = False) -> Self:
        return self.__class__.model_validate(self.model_dump())

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_computed_fields: bool = False,
        round_trip: bool = False,
        warnings: (bool | Literal["none", "warn", "error"]) = True,
        fallback: Callable[[Any], Any] | None = None,
        serialize_as_any: bool = False,
    ) -> dict[str, Any]:
        if mode not in VALID_MODES:
            raise ValueError
        if exclude_computed_fields:
            raise ValueError
        mode = cast(Literal["json", "python"], mode)

        return self.get_adapter().dump_python(
            self.pre_dump_hook(mode=mode),
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            fallback=fallback,
            serialize_as_any=serialize_as_any,
        )

    def model_dump_json(
        self,
        *,
        indent: int | None = None,
        ensure_ascii: bool = False,
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_computed_fields: bool = False,
        round_trip: bool = False,
        warnings: (bool | Literal["none", "warn", "error"]) = True,
        fallback: Callable[[Any], Any] | None = None,
        serialize_as_any: bool = False,
    ) -> str:
        if ensure_ascii or exclude_computed_fields:
            raise ValueError
        from_hook = self.pre_dump_hook(mode="json")
        return (
            self.get_adapter()
            .dump_json(
                from_hook,
                indent=indent,
                include=include,
                exclude=exclude,
                context=context,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                round_trip=round_trip,
                warnings=warnings,
                fallback=fallback,
                serialize_as_any=serialize_as_any,
            )
            .decode("utf-8")
        )

    @classmethod
    def model_json_schema(
        cls,
        by_alias: bool = True,
        ref_template: str = DEFAULT_REF_TEMPLATE,
        schema_generator: type[GenerateJsonSchema] = GenerateJsonSchema,
        mode: JsonSchemaMode = "validation",
        *,
        union_format: Literal["any_of", "primitive_type_array"] = "any_of",
    ) -> dict[str, Any]:
        return cls.get_adapter().json_schema(
            by_alias=by_alias,
            ref_template=ref_template,
            schema_generator=schema_generator,
            mode=mode,
        )

    @classmethod
    def model_parametrized_name(cls, params: tuple[type[Any], ...]) -> str:
        raise NotImplementedError

    @classmethod
    def model_validate(
        cls,
        obj: Any,
        *,
        strict: bool | None = None,
        extra: ExtraValues | None = None,
        from_attributes: bool | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        by_name: bool | None = None,
    ) -> Self:
        if extra:
            raise ValueError
        adapter = cls.get_adapter()
        obj = cls.pre_validation_hook_python(obj)
        obj = adapter.validate_python(
            obj,
            strict=strict,
            from_attributes=from_attributes,
            context=context,
            by_alias=by_alias,
            by_name=by_name,
        )
        validated = cls(obj)
        return cls.post_validation_hook(validated)

    @classmethod
    def model_validate_json(
        cls,
        json_data: str | bytes | bytearray,
        *,
        strict: bool | None = None,
        extra: ExtraValues | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        by_name: bool | None = None,
    ) -> Self:
        if extra:
            raise ValueError
        adapter = cls.get_adapter()
        json_string = cls.pre_validation_hook_json(json_data)
        raw_validated = adapter.validate_json(
            json_string,
            strict=strict,
            context=context,
            by_alias=by_alias,
            by_name=by_name,
        )
        validated = cls(raw_validated)
        return cls.post_validation_hook(validated)

    @classmethod
    def model_validate_strings(
        cls,
        obj: Any,
        *,
        strict: bool | None = None,
        extra: ExtraValues | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        by_name: bool | None = None,
    ) -> Self:
        return cls.model_validate(
            obj,
            strict=strict,
            extra=extra,
            context=context,
            by_alias=by_alias,
            by_name=by_name,
        )

    @property
    def __annotations__(self) -> dict[str, Any]:  # type: ignore
        return self.get_adapter().__annotations__

    @classmethod
    def read_json_file(cls, read_path: Path) -> Self:
        return cls.model_validate_json(read_path.read_text())

    def write_json_file(self, write_path: Path, by_alias: bool = True) -> Path:
        write_json(cast(JSONDict, self.model_dump(mode="json", by_alias=by_alias)), write_path)
        return write_path

    @staticmethod
    def pre_validation_hook_python(python_dict: dict) -> dict:
        return python_dict

    @staticmethod
    def pre_validation_hook_json(json_string: str | bytes | bytearray) -> str | bytes | bytearray:
        return json_string

    @staticmethod
    def post_validation_hook(validated: T) -> T:
        return validated

    def pre_dump_hook(self, *, mode: Mode) -> Self:
        if mode not in VALID_MODES:
            raise ValueError
        return self

    def post_dump_hook(self, dumped: T, *, mode: Mode) -> T:
        return dumped

    def schema(self) -> CoreSchema:
        return self.get_adapter().core_schema

    def schema_json(self) -> str:
        return json.dumps(self.model_json_schema())

    def _validate_other(self, other) -> Self:
        if not self.__class__ == other.__class__:
            other = self.__class__.model_validate(other)
        return other


class BaseDict[K, V](dict[K, V], AbstractCustom):
    """
    Dictionary type leveraging pydantic for validation and JSON serialization.
    """

    @classmethod
    def get_adapter(cls) -> TypeAdapter[dict[K, V]]:
        """Return a TypeAdapter for this subclass, preserving its key/value types."""
        for base in getattr(cls, "__orig_bases__", []):
            if isinstance(base, GenericAlias):
                key_type, val_type = get_args(base)
                return TypeAdapter(dict[key_type, val_type])  # type: ignore[valid-type]
        raise TypeError("Key and value types not found.")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{dict.__repr__(self)}"

    def __str__(self) -> str:
        raise NotImplementedError

    def __or__(self, other: dict[K, V]) -> "BaseDict[K, V]":  # type: ignore
        return self | self._validate_other(other)

    def update(self, other) -> None:
        super().update(self._validate_other(other))


class BaseList[T](list[T], AbstractCustom):
    """
    Dictionary type leveraging pydantic for validation and JSON serialization.
    """

    @classmethod
    def get_adapter(cls) -> TypeAdapter[list[T]]:
        """Return a TypeAdapter for this subclass, preserving its key/value types."""
        for base in getattr(cls, "__orig_bases__", []):
            if isinstance(base, GenericAlias):
                element_type = get_args(base)[0]
                return TypeAdapter(list[element_type])  # type: ignore[valid-type]
        raise TypeError("Key and value types not found.")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{list.__repr__(self)}"

    def __str__(self) -> str:
        raise NotImplementedError

    def extend(self, other: Iterable) -> None:
        super().extend(self._validate_other(other))

    def append(self, other: Any) -> None:
        self.extend([other])


class BaseSet[T](set[T], AbstractCustom):
    """
    Dictionary type leveraging pydantic for validation and JSON serialization.

    TODO: make dump as list if not all members hashable.
    """

    @classmethod
    def get_adapter(cls) -> TypeAdapter[set[T]]:
        """Return a TypeAdapter for this subclass, preserving its key/value types."""
        for base in getattr(cls, "__orig_bases__", []):
            if isinstance(base, GenericAlias):
                element_type = get_args(base)[0]
                return TypeAdapter(set[element_type])  # type: ignore[valid-type]
        raise TypeError("Key and value types not found.")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{set.__repr__(self)}"

    def __str__(self) -> str:
        raise NotImplementedError

    def __or__(self, other) -> "BaseSet[T]":  # type: ignore
        return self | self._validate_other(other)

    def add(self, other: T) -> None:
        self.update({other})

    def update(self, *s: Iterable[T]) -> None:
        for other in s:
            super().update(self._validate_other(other))

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: IncEx | None = None,
        exclude: IncEx | None = None,
        context: Any | None = None,
        by_alias: bool | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_computed_fields: bool = False,
        round_trip: bool = False,
        warnings: (bool | Literal["none", "warn", "error"]) = True,
        fallback: Callable[[Any], Any] | None = None,
        serialize_as_any: bool = False,
    ) -> dict[str, Any]:
        if mode not in VALID_MODES:
            raise ValueError
        if exclude_computed_fields:
            raise ValueError

        return BaseList.model_validate(self).model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
            fallback=fallback,
            serialize_as_any=serialize_as_any,
        )



import re
from collections.abc import Hashable, Mapping, Sequence
from typing import Annotated, TypeVar

import annotated_types
from pydantic import NonNegativeInt

from .collections import Immutable
from .typing import Re

__all__ = (
    "AlphaClosed",
    "AlphaOpen",
    "DateString",
    "DateStringFlex",
    "DateStringFlex",
    "H",
    "Immutable",
    "JsonArray",
    "JsonContainer",
    "JsonObject",
    "JsonValue",
    "Natural",
    "NegativeScore",
    "PolarityScore",
    "Positive",
    "PositiveScore",
    "Stars",
    "T",
    "TimestampISO",
    "Timestamp_",
)

T = TypeVar("T")
H = TypeVar("H", bound=Hashable)

type Natural = NonNegativeInt
type Stars = Annotated[float, annotated_types.Ge(1.0), annotated_types.Le(5.0)]
type Positive = Annotated[float, annotated_types.Gt(0)]
type PositiveScore = Annotated[float, annotated_types.Ge(0.0), annotated_types.Le(5.0)]
type NegativeScore = Annotated[float, annotated_types.Ge(-5.0), annotated_types.Le(0.0)]
type AlphaClosed = Annotated[float, annotated_types.Ge(0.0), annotated_types.Le(1.0)]
type AlphaOpen = Annotated[float, annotated_types.Gt(0.0), annotated_types.Lt(1.0)]
type PolarityScore = Annotated[float, annotated_types.Ge(-1.0), annotated_types.Le(1.0)]
type DateString = Annotated[str, Re(pattern=re.compile(r"^[12]\d{3}-\d{2}-\d{2}$"))]
type DateStringFlex = Annotated[str, Re(pattern=re.compile(r"^[12]\d{3}-\d{1,2}-\d{1,2}$"))]
type TimestampISO = Annotated[
    str, Re(pattern=re.compile(r"^[12]\d{3}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}"))
]
type Timestamp_ = Annotated[
    str, Re(pattern=re.compile(r"^[12]\d{3}-\d{2}-\d{2}_\d{2}:\d{2}:\d{2}"))
]

type JsonValue = int | float | str | bool | None
type JsonArray = Sequence[JsonValue | JsonObject]
type JsonObject = Mapping[str, JsonValue | JsonObject | JsonArray]
type JsonContainer = JsonArray | JsonObject



__all__ = ("Colorizer", "color", "nullcolor")


import re


class Colorizer:
    def __init__(self, use_colors: bool = True) -> None:
        if use_colors:
            self.BLACK = "\u001b[30m"
            self.RED = "\u001b[31m"
            self.GREEN = "\u001b[32m"
            self.YELLOW = "\u001b[33m"
            self.BLUE = "\u001b[34m"
            self.MAGENTA = "\u001b[35m"
            self.CYAN = "\u001b[36m"
            self.WHITE = "\u001b[37m"
            self.RESET = "\u001b[0m"
        else:
            self.BLACK = ""
            self.RED = ""
            self.GREEN = ""
            self.YELLOW = ""
            self.BLUE = ""
            self.MAGENTA = ""
            self.CYAN = ""
            self.WHITE = ""
            self.RESET = ""

    @staticmethod
    def strip(s: str) -> str:
        # OLD return s[5:-4] if s.startswith("\u001b") else s
        ansi_escape = re.compile(r"\x1b\[[0-9;]*[a-zA-Z]")
        return ansi_escape.sub("", s)

    @staticmethod
    def length(s: str) -> int:
        return len(Colorizer.strip(s))

    def black(self, text: object) -> str:
        return self._format(text, self.BLACK)

    def red(self, text: object) -> str:
        return self._format(text, self.RED)

    def green(self, text: object) -> str:
        return self._format(text, self.GREEN)

    def yellow(self, text: object) -> str:
        return self._format(text, self.YELLOW)

    def blue(self, text: object) -> str:
        return self._format(text, self.BLUE)

    def magenta(self, text: object) -> str:
        return self._format(text, self.MAGENTA)

    def cyan(self, text: object) -> str:
        return self._format(text, self.CYAN)

    def white(self, text: object) -> str:
        return self._format(text, self.WHITE)

    def _format(self, text: object, color_code: str) -> str:
        return f"{color_code}{text!s}{self.RESET}"


color = Colorizer()
nullcolor = Colorizer(use_colors=False)



from collections.abc import Callable
from operator import not_ as negate

__all__ = (
    "as_bool",
    "equal_within",
    "make_maybe_not",
    "match_empty",
    "nearly_equal",
    "trivial_condition",
)


def nearly_equal(a: int | float, b: int | float) -> bool:
    return abs(a - b) < 0.00001


def equal_within(a: int | float, b: int | float, epsilon: float | int) -> bool:
    return abs(a - b) < epsilon


def match_empty(item: str | None | bool | set[str] | float) -> bool:
    return not bool(item)


def trivial_condition(n: object) -> bool:
    return True


def as_bool(object_: object) -> bool:
    if isinstance(object_, bool):
        return object_
    if isinstance(object_, int | float):
        return True
    return bool(object_)


def make_maybe_not(*, negated: bool) -> Callable[[object], bool]:
    return negate if negated else as_bool



from collections.abc import Callable
from enum import StrEnum
from enum import auto as _auto
from typing import cast

auto = cast(Callable[[], str], _auto)


class StrEnumUpper(StrEnum):
    """ """

    @staticmethod
    def _generate_next_value_(name: str, start, count, last_values) -> str:  # type: ignore
        return name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}"



import importlib
import sys
from functools import cache
from importlib.util import find_spec
from types import ModuleType
from typing import Any, ClassVar, cast

from loguru import logger


def _make_message(mod_name: str, error: Exception) -> str:
    return f"Attempted import of dependency '{mod_name}' unsuccessful ({error.__class__}): {error}"


class LazyModule(ModuleType):
    """A proxy that imports a module lazily upon first attribute access."""

    _SAFE_ATTRS: ClassVar[set[str]] = {
        "__name__",
        "__loader__",
        "__package__",
        "__spec__",
        "__path__",
        "__file__",
        "_module",
        "_load",
    }

    def __init__(self, name: str):
        super().__init__(name)
        self._module: ModuleType | None = None

    def _load(self) -> ModuleType:
        if self._module is None:
            _name = self.__name__
            sys.modules.pop(_name, None)
            module = importlib.import_module(_name)
            sys.modules[_name] = module
            setattr(self, "_module", module)
        return cast(ModuleType, self._module)

    def __getattribute__(self, attr: str) -> Any:
        if attr in object.__getattribute__(self, "_SAFE_ATTRS"):
            return object.__getattribute__(self, attr)

        module = (
            object.__getattribute__(self, "_module") or object.__getattribute__(self, "_load")()
        )
        return getattr(module, attr)

    def __dir__(self) -> list[str]:
        self._load()
        return dir(self._module)

    def __repr__(self):
        return f"<LazyModule proxy for '{self.__name__}'>"


@cache
def maybe_import_lazy(module_name: str) -> ModuleType | None:
    """
    Try to lazily import a module by name. Returns a proxy that loads on first use,
        or None if unavailable.

    Results are cached for efficiency.
    """
    if module_name in sys.modules:
        return sys.modules[module_name]

    try:
        spec = find_spec(module_name)
        if spec is None:
            return None

        proxy = LazyModule(module_name)
        sys.modules[module_name] = proxy
        return proxy
    except (ModuleNotFoundError, ImportError) as e:
        logger.debug(_make_message(module_name, e))
        return None


@cache
def import_lazy(module_name: str) -> ModuleType:
    """
    Lazily import a module by name. Returns a proxy that loads on first use,
        or None if unavailable.

    Results are cached for efficiency.
    """
    if module_name in sys.modules:
        return sys.modules[module_name]

    try:
        spec = find_spec(module_name)
        if spec is None:
            raise ModuleNotFoundError

        proxy = LazyModule(module_name)
        sys.modules[module_name] = proxy
        return proxy
    except (ModuleNotFoundError, ImportError) as e:
        logger.debug(_make_message(module_name, e))
        raise (e)


@cache
def maybe_import(module_name: str) -> ModuleType | None:
    try:
        return importlib.import_module(module_name)
    except ModuleNotFoundError as e:
        logger.debug(_make_message(module_name, e))
        return None



from collections.abc import Callable
from functools import wraps
from typing import TypeVar, cast

# from .functional import identity


T = TypeVar("T")
In = TypeVar("In")
Out = TypeVar("Out")


def trivial_decorator[In, Out](func: Callable[[In], Out]) -> Callable[[In], Out]:
    return func


def impure[In, Out](
    callable_or_none: Callable[[In], Out] | None = None,
    message: str = "",
) -> Callable[[In], Out] | Callable[[Callable[[In], Out]], Callable[[In], Out]]:
    if callable_or_none is None:
        return trivial_decorator
    else:
        return callable_or_none


def pure[In, Out](
    callable_or_none: Callable[[In], Out] | None = None,
    message: str = "",
) -> Callable[[In], Out] | Callable[[Callable[[In], Out]], Callable[[In], Out]]:
    if callable_or_none is None:
        return trivial_decorator
    else:
        return callable_or_none


def endo(
    callable_or_none: Callable[[T], T] | None = None,
    message: str = "",
) -> Callable[[T], T] | Callable[[Callable[[In], Out]], Callable[[In], Out]]:
    if callable_or_none is None:
        return trivial_decorator
    else:
        return callable_or_none


def decorate_with_message[In, Out](func: Callable[[In], Out], message: str) -> Callable[[In], Out]:
    @wraps(func)
    def wrapper(*fargs, **fkwargs) -> Out:  # type: ignore
        print(message)

        return func(*fargs, **fkwargs)  # type: ignore

    return cast(Callable[[In], Out], wrapper)


def mutates_instance[In, Out](func: Callable[[In], Out], *mutated: str) -> Callable[[In], Out]:
    message = (
        f"\u001b[31mMutating in place\u001b[0m:         "
        f" via \u001b[33m{func.__name__:<25}\u001b[0m"
        f" in \u001b[34m{func.__module__}\u001b[0m"
    )

    return decorate_with_message(func, message)


def mutates_and_returns_instance[In, Out](
    func: Callable[[In], Out],
    *mutated: str,
) -> Callable[[In], Out]:
    message = (
        f"\u001b[31mMutating\u001b[0m"
        f"{(' ' * bool(mutated)) + ', '.join(mutated)}"
        f" \u001b[31mand returning instance of\u001b[0m: {func.__class__.__name__}"
        f" via \u001b[33m{func.__name__:<25}\u001b[0m"
        f" in \u001b[34m{func.__module__}\u001b[0m"
    )

    return decorate_with_message(func, message)


def mutates[In, Out](func: Callable[[In], Out], *mutated: str) -> Callable[[In], Out]:
    message = (
        f"Mutating \u001b[36m{', '.join(mutated):<50}\u001b[0m"
        f" via \u001b[33m{func.__name__:<25}\u001b[0m"
        f" in \u001b[34m{func.__module__:<40}\u001b[0m"
    )

    return decorate_with_message(func, message)


def refactor[In, Out](func: Callable[[In], Out], message: str = "") -> Callable[[In], Out]:
    message = (
        f"\u001b[36mREFACTOR\u001b[0m"
        f" \u001b[33m{func.__name__}\u001b[0m"
        f" in \u001b[34m{func.__module__.replace('consilium.', '.')}\u001b[0m."
        f"\n    {message}"
    )

    return decorate_with_message(func, message)



import subprocess
import sys
from pathlib import Path
from typing import cast

from .color import Colorizer
from .fp import lmap

type CommandList = list[str | Path | int | float] | list[str]


color = Colorizer()


def resolve_variables(
    commands: CommandList, env: dict[str, str | float | int | Path]
) -> CommandList:
    def get_var(s: str) -> str | float | int | Path:
        if s.startswith("$"):
            print(env.get(s[1:], ""))
            return env.get(s[1:], "")
        return s

    print(lmap(get_var, commands))
    return lmap(get_var, commands)  # type: ignore


def run(commands: CommandList, **kwargs) -> subprocess.CompletedProcess[bytes]:
    _kwargs = cast(dict, dict(check=True, capture_output=True) | kwargs)
    if "resolve" in _kwargs:
        del _kwargs["resolve"]
    return subprocess.run(list(map(str, commands)), **_kwargs)  # noqa: PLW1510


def capture(commands: CommandList, **kwargs) -> str:
    _commands: CommandList = commands
    _kwargs = cast(dict, dict(capture_output=True, check=False) | kwargs)
    if "resolve" in _kwargs:
        del _kwargs["resolve"]
        _commands = resolve_variables(commands, _kwargs.get("env", {}))
    return subprocess.run(list(map(str, _commands)), **_kwargs).stdout.decode().strip()  # noqa: PLW1510


def run_with_result(
    commands: CommandList,
    print_output: bool = False,
    fail_on_error: bool = True,
    **kwargs,
) -> tuple[bool, str]:
    _commands: CommandList = commands
    _kwargs = dict(check=True, capture_output=True) | kwargs  # type: ignore
    if "resolve" in _kwargs:
        del _kwargs["resolve"]
        _commands = resolve_variables(commands, _kwargs.get("env", {}))  # type: ignore
    try:
        subproc_result = subprocess.run(list(map(str, _commands)), **_kwargs)  # type: ignore  # noqa: PLW1510
        result = not bool(subproc_result.returncode)
        stdout = subproc_result.stdout.decode().strip()
        stdout = stdout + "\n" * bool(stdout)
        stderr = subproc_result.stderr.decode().strip()
        stderr = stderr + "\n" * bool(stdout)
    except Exception as e:
        stdout = ""
        stderr = str(e)
        result = False

    bar = "-" * 36
    topbar = f"{'=' * 36} {color.blue('COMMAND')} {'=' * 35}"
    command_literal = " ".join(list(map(str, _commands)))
    msg = (
        f"\n{topbar}\n{color.blue(command_literal)}"
        f"\n\n{bar} {color.green('STDOUT')} {bar}\n{stdout}\n{bar}"
        f" {color.red('STDERR')} {bar}\n{stderr}\n"
    )
    if fail_on_error and not result:
        print(color.red(f"\nCommand failed: {command_literal}'"))
        print(msg)
        sys.exit(1)
    if print_output:
        print()
        print(msg)
    return result, msg



""" """

from collections.abc import Callable, Iterable, Sequence
from typing import (
    TYPE_CHECKING,
    TypeVar,
    TypeVarTuple,
    cast,
    overload,
)

from .fp import identity

if TYPE_CHECKING:
    from _typeshed import SupportsRichComparison  # pragma: no cover
else:
    type SupportsRichComparison = float

S = TypeVar("S")
T = TypeVar("T")
Tup = TypeVar("Tup", bound=tuple)
Ts = TypeVarTuple("Ts")
C = TypeVar("C", bound=SupportsRichComparison)


@overload
def keep_first[T, *Ts](tups: list[tuple[T, *Ts]]) -> list[T]: ...
@overload
def keep_first[T, *Ts](tups: tuple[tuple[T, *Ts], ...]) -> tuple[T, ...]: ...
def keep_first[T, *Ts](tups: Iterable[tuple[T, *Ts]]) -> Sequence[T]:
    if isinstance(tups, list):
        return [t[0] for t in tups]
    return tuple(t[0] for t in tups)


@overload
def keep_second[S, T, *Ts](tups: tuple[tuple[S, T, *Ts], ...]) -> tuple[T, ...]: ...
@overload
def keep_second[S, T, *Ts](tups: list[tuple[S, T, *Ts]]) -> list[T]: ...
def keep_second[S, T, *Ts](tups: Iterable[tuple[S, T, *Ts]]) -> Sequence[T]:
    if isinstance(tups, list):
        return [t[1] for t in tups]
    return tuple(t[1] for t in tups)


@overload
def keep_last[*Ts, T](tups: tuple[tuple[*Ts, T], ...]) -> tuple[T, ...]: ...
@overload
def keep_last[*Ts, T](tups: list[tuple[*Ts, T]]) -> list[T]: ...
def keep_last[*Ts, T](tups: Iterable[tuple[*Ts, T]]) -> Sequence[T]:
    if isinstance(tups, list):
        return [t[-1] for t in tups]
    return tuple(t[-1] for t in tups)


def _make_selector[T](
    u: int | slice | tuple[int, ...] | None = None,
) -> Callable[[tuple[T, ...]], T]:
    u = u if u is not None else slice(None)

    if isinstance(u, tuple):

        def _use(t: tuple[object, ...]) -> T:  # type: ignore
            return cast(T, tuple(t[i] for i in u))

    else:

        def _use(t: tuple[object, ...]) -> T:  # type: ignore
            return cast(T, t[u])

    return _use


@overload
def filter_tuples[Tup](
    seq: tuple[Tup, ...],
    *,
    use: int | slice | tuple[int, ...] | None = None,
    condition: Callable[[object], bool] = bool,
) -> tuple[Tup, ...]: ...


@overload
def filter_tuples[Tup, T](
    seq: list[Tup],
    *,
    use: int | slice | tuple[int, ...] | None = None,
    condition: Callable[[T], bool] = bool,
) -> Sequence[Tup]: ...


def filter_tuples[Tup, T](
    seq: Iterable[Tup],
    *,
    use: int | slice | tuple[int, ...] | None = None,
    condition: Callable[[T], bool] = bool,
) -> Sequence[Tup]:
    use = use if use is not None else slice(use)
    select = _make_selector(use)

    def _condition(elem: Tup) -> bool:
        return condition(select(cast(tuple[T, ...], elem)))

    filtered = filter(_condition, seq)
    if isinstance(seq, list):
        return list(filtered)
    return tuple(filtered)


@overload
def sort_tuples[Tup](
    seq: list[Tup],
    *,
    use: int | slice | tuple[int, ...] | None = None,
    key: Callable[[Tup], SupportsRichComparison] | None = None,
) -> list[Tup]: ...


@overload
def sort_tuples[Tup](
    seq: tuple[Tup, ...],
    *,
    use: int | slice | tuple[int, ...] | None = None,
    key: Callable[[Tup], SupportsRichComparison] | None = None,
) -> tuple[Tup, ...]: ...


def sort_tuples[Tup](
    seq: Iterable[Tup],
    *,
    use: int | slice | tuple[int, ...] | None = None,
    key: Callable[[Tup], SupportsRichComparison] | None = None,
) -> Sequence[Tup]:
    select = _make_selector(use)

    key = key if key is not None else cast(Callable[[Tup], SupportsRichComparison], identity)

    def _key(elem: Tup) -> SupportsRichComparison:
        selected = cast(Tup, select(cast(tuple[object, ...], elem)))
        return key(selected)

    return (list if isinstance(seq, list) else tuple)(sorted(seq, key=_key))



import builtins
import importlib
from types import ModuleType


def try_import(name: str, alias: str | None = None) -> ModuleType | None:
    try:
        mod = importlib.import_module(alias or name)
        setattr(builtins, name, mod)
    except ImportError:
        msg = (
            f"Could not import '{name}' (attempted import as '{alias}')."
            if alias
            else f"Could not import module '{name}'."
        )
        print(msg)
    return None


setattr(builtins, "try_import", try_import)


def import_ds(  # noqa: C901 - allow 'too complex'
    np: bool = True,
    pd: bool = True,
    pl: bool = False,
    plt: bool = True,
    sns: bool = False,
    px: bool = False,
    scipy: bool = False,
    sklearn: bool = False,
    sm: bool = False,
    torch: bool = False,
    tf: bool = False,
    transformers: bool = False,
) -> None:
    if np:
        try_import("numpy", alias="np")
    if pd:
        try_import("pandas", alias="pd")
    if pl:
        try_import("polars", alias="pl")
    if plt:
        try_import("matplotlib.pyplot", alias="plt")
    if sns:
        try_import("seaborn", alias="sns")
    if px:
        try_import("plotly.express", alias="px")
    if scipy:
        try_import("scipy")
    if sm:
        try_import("statsmodels", alias="sm")
    if sklearn:
        try_import("sklearn")
    if torch:
        try_import("torch")
    if tf:
        try_import("tensorflow", alias="tf")
    if transformers:
        try_import("transformers")



from collections.abc import Callable
from random import choice, choices, randint, random
from string import ascii_letters, ascii_lowercase, ascii_uppercase
from typing import Literal


def pascalcase(
        subwords: int | None = None,
        max_length: int | None = None,
        sub_length_min: int = 3,
        sub_length_max: int = 6,
    ) -> str:
        if not max_length:
            max_length = randint(5, 12)
        if not subwords:
            subwords = int(max_length / sub_length_min) + 1
        return "".join(
            titlecase(randint(sub_length_min, sub_length_max)) for _ in range(subwords)
        )[:max_length]

def camelcase(
        subwords: int | None = None,
        max_length: int | None = None,
        sub_length_min: int = 3,
        sub_length_max: int = 6,
    ) -> str:
        if not max_length:
            max_length = randint(5, 12)
        if not subwords:
            subwords = int(max_length / sub_length_min) + 1
        return "".join(
            (
                lowercase(randint(sub_length_min, sub_length_max)),
                *(titlecase(randint(sub_length_min, sub_length_max)) for _ in range(subwords)),
            )
        )[:max_length]

def snakecase(
        subwords: int | None = None,
        max_length: int | None = None,
        sub_length_min: int = 3,
        sub_length_max: int = 6,
    ) -> str:
        if not max_length:
            max_length = randint(5, 12)
        if not subwords:
            subwords = int(max_length / sub_length_min) + 1
        return "_".join(
            lowercase(randint(sub_length_min, sub_length_max)) for _ in range(subwords)
        )[:max_length]

def kebabcase(
        subwords: int | None = None,
        max_length: int | None = None,
        sub_length_min: int = 3,
        sub_length_max: int = 6,
    ) -> str:
        if not max_length:
            max_length = randint(5, 12)
        if not subwords:
            subwords = int(max_length / sub_length_min) + 1
        return "-".join(
            lowercase(randint(sub_length_min, sub_length_max)) for _ in range(subwords)
        )[:max_length]

def string_tuple(
        string_type: Literal["any", "lower", "upper", "camel", "pascal"] = "any",
        length: int | None = None,
        string_length: int | None = None,
    ) -> tuple[str, ...]:
        stringmaker: Callable[[int | None], str]
        stringmakers: dict[str, Callable[[int | None], str]] = {
            "any": string,
            "lower": lowercase,
            "upper": uppercase,
            "camel": camelcase,
            "pascal": pascalcase,
            "snake": snakecase,
            "kebab": kebabcase,
        }
        stringmaker = stringmakers[string_type]

        if not length:
            length = randint(2, 9)
        return tuple(stringmaker(string_length) for i in range(length))

def text(num_words: int = 10, min_word_length: int = 3, max_word_length: int = 10) -> str:
        return " ".join(
            string(length=randint(min_word_length, max_word_length)) for _ in range(num_words)
        )

def string(length: int | None = None) -> str:
        if not length:
            length = randint(2, 9)
        return "".join(choices(ascii_letters, k=length))

def lowercase(length: int | None = None) -> str:
        if not length:
            length = randint(2, 9)
        return "".join(choices(ascii_lowercase, k=length))

def titlecase(length: int | None = None) -> str:
        if not length:
            length = randint(2, 9)
        return choice(ascii_uppercase) + "".join(choices(ascii_lowercase, k=length - 1))

def uppercase(length: int | None = None) -> str:
        if not length:
            length = randint(2, 9)
        return "".join(choices(ascii_uppercase, k=length))

def int_tuple(
        size: int | None = None,
        min_val: int = 0,
        max_val: int = 20,
    ) -> tuple[int, ...]:
        if size is None:
            size = randint(2, 10)
        return tuple(randint(min_val, max_val) for _ in range(size))

def randfloat(lower: float = 0.0, upper: float = 1.0) -> float:
        return lower + random() * (upper - lower)


if __name__ == "__main__":
    print(".string():       ", string())
    print(".lowercase():    ", lowercase())
    print(".titlecase():    ", titlecase())
    print(".uppercase():    ", uppercase())
    print(".pascalcase():   ", pascalcase())
    print(".camelcase():    ", camelcase())
    print(".snakecase():    ", snakecase())
    print(".kebabcase():    ", kebabcase())
    print(".string_tuple():  ", string_tuple())
    print(".int_tuple():    ", int_tuple())
    print(".randfloat():    ", randfloat())
    print()
    print(".string():       ", string())
    print(".lowercase():    ", lowercase())
    print(".titlecase():    ", titlecase())
    print(".uppercase():    ", uppercase())
    print(".pascalcase():   ", pascalcase())
    print(".camelcase():    ", camelcase())
    print(".snakecase():    ", snakecase())
    print(".kebabcase():    ", kebabcase())
    print(".string_tuple():  ", string_tuple("pascal"))
    print(".int_tuple():    ", int_tuple(size=20, min_val=50, max_val=60))
    print(".randfloat():    ", randfloat(6.0, 7.0))



import json
import os
import time
from collections.abc import Callable
from datetime import datetime
from pathlib import Path
from typing import Any, TypeVar


def DEFAULT_PATH() -> Path:
    return Path("codeqa/performance/perf_log/perf_log.jsonl")


_I = TypeVar("_I")
_O = TypeVar("_O")


def get_callback_name(cb: Callable) -> str:
    return getattr(cb, "__name__", None) or cb.__class__.__name__


def log_perf(
    callable_or_none: Callable | None = None,
    *,
    log_path: Path | str | None | Callable[[], Path] = None,
    extra_info: dict[str, Any] | None = None,
) -> Callable:
    def wrapper(_callback: Callable[[_I], _O]) -> Callable[[_I], _O]:
        if not os.environ.get("LOG_PERFORMANCE"):
            return _callback

        def inner(*args, **kwargs) -> _O:
            start_time = time.time()
            result = _callback(*args, **kwargs)
            end_time = time.time()

            log_info = (
                {
                    "name": get_callback_name(_callback),
                    "timeElapsed": round(end_time - start_time, 4),
                    "timestamp": datetime.now().isoformat(sep="_")[:-3],
                    "file": _callback.__module__,
                }
                | (extra_info or {})
                | {"args": str(args), "kwargs": str(kwargs)}
            )
            _log_path: Path = (
                Path(log_path)
                if isinstance(log_path, Path | str)
                else (log_path() if callable(log_path) else DEFAULT_PATH())
            )
            with _log_path.open("a", encoding="utf-8") as f:
                f.write(json.dumps(log_info) + "\n")

            return result

        return inner

    if callable(callable_or_none):
        return wrapper(callable_or_none)
    return wrapper



import textwrap
from collections import Counter, defaultdict
from collections.abc import Callable, Hashable, Iterable
from pathlib import Path

from .color import color, nullcolor
from .fp import identity

__all__ = ("objtree", "print_tree", "wrap_line")


def freq(
    things: Iterable[Hashable],
    min: int = 1,
    max: int | None = None,
    sep: str = "  ",
    use_color: bool = True,
) -> None:
    c = Counter(things)
    magenta = (color if use_color else nullcolor).magenta
    max_ = max or c.total()
    places = len(str(max_))
    for thing, count in c.most_common():
        if min <= count <= max_:
            print(f"{magenta(str(count)):>{places}}{sep}{thing!s}")


def print_tree(strings):
    def tree() -> defaultdict:
        return defaultdict(tree)

    root = tree()

    for _string in strings:
        parts = _string.split(".")
        current_level = root
        for part in parts:
            current_level = current_level[part]

    def print_subtree(node, prefix=""):
        children = list(node.keys())
        for i, child in enumerate(children):
            is_last = i == len(children) - 1
            if is_last:
                print(prefix + "└─ " + child)
                new_prefix = prefix + "   "
            else:
                print(prefix + "├─ " + child)
                new_prefix = prefix + "│  "
            print_subtree(node[child], new_prefix)

    print_subtree(root)


def wrap_line(line: str, length: int, formatter: Callable[[str], str] = identity) -> str:
    wrapped = textwrap.wrap(line, width=length)
    return "\n".join(formatter(line) for line in wrapped)


_LEAVES: tuple[type, ...] = (str, int, float, bool, type(None), bytes, Path)


def build_objtree(
    obj,
    name=None,
    max_depth: int = 4,
    indent: str = "",
    _depth: int = 0,
    _seen: set[int] | None = None,
    _lines: list[str] | None = None,
) -> list[str]:
    if _seen is None:
        _seen = set()
    if _lines is None:
        _lines = []

    name = name or getattr(obj, "__name__", None) or f"<{type(obj).__name__}>"

    leader = "    " if _depth > 0 else ""  # formerly " └─ "
    type_name = color.blue(type(obj).__name__)
    _lines.append(f"{indent}{leader}{color.magenta(name)} ({type_name})")

    obj_id = id(obj)
    if _depth >= max_depth or obj_id in _seen or isinstance(obj, _LEAVES):
        return _lines

    _seen.add(obj_id)

    # Get attributes, ignoring dunders and private-ish members
    attrs = sorted([a for a in dir(obj) if not a.startswith("__")])

    for attr in attrs:
        try:
            val = getattr(obj, attr)
            build_objtree(
                val,
                name=attr,
                max_depth=max_depth,
                indent=indent + ("    " if _depth > 0 else ""),
                _depth=_depth + 1,
                _seen=_seen,
                _lines=_lines,
            )
        except Exception:  # pragma: no cover
            continue

    return _lines


# def build_objtree(
#     obj,
#     name=None,
#     max_depth: int = 4,
#     indent: str = "",
#     leader: str = "    ",
#     _depth: int = 0,
#     _seen: set[int] | None = None,
#     _lines: list[str] | None = None,
# ) -> list[str]:
#     if _seen is None:
#         _seen = set()
#     if _lines is None:
#         _lines = []

#     _leader = leader if _depth else ""
#     if not name:
#         if hasattr(obj, "__name"):
#             name = obj.__name__
#         elif hasattr(obj, "__repr__"):
#             name = repr(obj)
#         else:
#             name = "<root>"

#     type_name = color.blue(type(obj).__name__)
#     _lines.append(f"{indent}{_leader}{color.magenta(name)} ({type_name})")

#     obj_id = id(obj)
#     if (
#         _depth >= max_depth
#         or obj_id in _seen
#         or isinstance(obj, (str, int, float, bool, type(None)))
#     ):
#         return _lines

#     _seen.add(obj_id)
#     attrs = [a for a in dir(obj) if not a.startswith("__")]

#     for attr in attrs:
#         try:
#             val = getattr(obj, attr)
#             build_objtree(
#                 val,
#                 name=attr,
#                 max_depth=max_depth,
#                 indent=indent + "    ",
#                 leader=_leader,
#                 _depth=_depth + 1,
#                 _seen=_seen,
#                 _lines=_lines,
#             )
#         except Exception:
#             continue

#     return _lines


def objtree(ob: object) -> str:
    return "\n".join(build_objtree(ob))


def otree(ob: object) -> None:
    print(objtree(ob))  # pragma: no cover



from collections.abc import Callable, Hashable, Iterable
from operator import attrgetter, itemgetter
from typing import TypeVar, cast

T = TypeVar("T")
H = TypeVar("H", bound=Hashable)


def dep_sort(  # noqa: C901 (allow complex function)
    objects: Iterable[T],
    *,
    namegetter: str | Callable[[T], H],
    depgetter: str | Callable[[T], Iterable[H]],
    verbose: bool = False,
) -> list[T]:
    """
    Sort list of objects such that dependency constraints are satisfied.
    """
    objects = list(objects)
    if not objects:
        return []
    getter: type[itemgetter] | type[attrgetter] = (
        itemgetter if isinstance(objects[0], dict) else attrgetter
    )
    _namegetter = cast(
        Callable[[T], H], getter(namegetter) if isinstance(namegetter, str) else namegetter
    )
    _depgetter = cast(
        Callable[[T], Iterable[H]], getter(depgetter) if isinstance(depgetter, str) else depgetter
    )

    def safe_depgetter(ob: T) -> tuple[H, ...]:
        try:
            return tuple(_depgetter(ob))
        except Exception:
            return tuple()

    registry: dict[H, T] = {_namegetter(t): t for t in objects}
    ids = list(registry)
    dep_dict: dict[H, tuple[H, ...]] = {h: safe_depgetter(t) for h, t in registry.items()}

    def place_after(to_move: H, deps: tuple[H, ...], id_list: list) -> bool:
        if verbose:
            print(f"\nto_move: {to_move}; deps: {','.join(map(str, deps))}")
            print(f"Before:  {id_list}")
        if not deps:
            return False
        idx1 = id_list.index(to_move)
        idx2 = max(map(id_list.index, deps))
        if idx1 > idx2:
            return False
        id_list.pop(idx1)
        id_list.insert(idx2, to_move)
        if verbose:
            print(f"idx1: {idx1}, idx2: {idx2}")
            print(f"After:  {id_list}")
        return True

    print(ids)
    count = 0
    maxcount = sum(range(len(ids) + 1))
    while count < maxcount:
        change_tracker = []
        for _id, _deps in dep_dict.items():
            changed = place_after(_id, _deps, ids)
            change_tracker.append(changed)
        unchanged = not any(change_tracker)
        print(ids)
        if unchanged:
            return [registry[i] for i in ids]
        count += 1

    raise ValueError("Graph contains a cycle.")



import re
from abc import ABC, abstractmethod
from collections.abc import Callable, Generator, Hashable, Iterable, KeysView, ValuesView
from dataclasses import dataclass
from pathlib import Path
from types import UnionType
from typing import Any, Protocol, Self, TypeVar, cast

from pydantic_core import core_schema

T = TypeVar("T")
H = TypeVar("H", bound=Hashable)


type Atomic = int | float | str | bool | None
type ClassInfo = type | UnionType | tuple["ClassInfo"]
type JSONPrimitive = int | float | str | None
type JSONDict = dict[str, JSONPrimitive | "JSONDict" | list["JSONDict | JSONPrimitive"]]
# alternative: JSONDict = dict[str, Union["JSONDict", str, list[str], tuple[str, ...], set[str]]]
type JSONList = list[JSONPrimitive | JSONDict]
type Endofunction[T] = Callable[[T], T]
type SequenceAlias[T] = (
    list[T]
    | tuple[T]
    | set[T]
    | map[T]
    | filter[T]
    | dict[T, Any]
    | KeysView[T]
    | ValuesView[T]
    | range
    | Generator[T]
)
_SequenceAliasVar = (
    list | tuple | set | map | filter | dict | KeysView | ValuesView | range | Generator
)

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_T_contra = TypeVar("_T_contra", contravariant=True)


def areinstances(iterable_instance: Iterable[Any], class_or_tuple: ClassInfo) -> bool:
    return all(map(lambda inst: isinstance(inst, class_or_tuple), iterable_instance))


def fallback_if_none(orig: T | None, alt: T) -> T:
    return alt if (orig is None) else orig


def call_fallback_if_none(orig: T | None, alt: Callable[[], T]) -> T:
    return alt() if (orig is None) else orig


def call_if_callable(maybe_callable: T | Callable[[], T]) -> T:
    if hasattr(maybe_callable, "__call__"):
        return cast(Callable[[], T], maybe_callable)()
    return cast(T, maybe_callable)


maybe_call = call_if_callable


class SupportsGe(Protocol):
    def __ge__(self: T, __other: T) -> bool: ...


class SupportsGt(Protocol):
    def __gt__(self: T, __other: T) -> bool: ...


class SupportsLe(Protocol):
    def __le__(self: T, __other: T) -> bool: ...


class SupportsLt(Protocol):
    def __lt__(self: T, __other: T) -> bool: ...


@dataclass(frozen=True)
class _BaseMetadata:
    """Base class for all metadata.

    This exists mainly so that implementers
    can do `isinstance(..., BaseMetadata)` while traversing field annotations.
    """

    __slots__ = ()


@dataclass(frozen=True)
class Ge(_BaseMetadata):
    """Ge(ge=x) implies that the value must be greater than or equal to x.

    It can be used with any type that supports the ``>=`` operator,
    including numbers, dates and times, strings, sets, and so on.
    """

    ge: SupportsGe

    def __get_pydantic_core_schema__(
        self, source_type, _handler
    ) -> core_schema.WrapValidatorFunctionSchema:
        string_schema = core_schema.any_schema()

        return core_schema.no_info_wrap_validator_function(
            cast(core_schema.NoInfoWrapValidatorFunction, self._validate), string_schema
        )

    def _validate(
        self, value: SupportsGe, handler: core_schema.ValidatorFunctionWrapHandler
    ) -> SupportsGe:
        if not value >= self.ge:
            raise ValueError(f"Value {value} is not less than or equal to {self.ge}.")
        return value


@dataclass(frozen=True)
class Gt(_BaseMetadata):
    """Gt(gt=x) implies that the value must be greater than x.

    It can be used with any type that supports the ``>`` operator,
    including numbers, dates and times, strings, sets, and so on.
    """

    gt: SupportsGt

    def __get_pydantic_core_schema__(
        self, source_type, _handler
    ) -> core_schema.WrapValidatorFunctionSchema:
        string_schema = core_schema.any_schema()

        return core_schema.no_info_wrap_validator_function(
            cast(core_schema.NoInfoWrapValidatorFunction, self._validate), string_schema
        )

    def _validate(
        self, value: SupportsGt, handler: core_schema.ValidatorFunctionWrapHandler
    ) -> SupportsGt:
        if not value > self.gt:
            raise ValueError(f"Value {value} is not less than or equal to {self.gt}.")
        return value


@dataclass(frozen=True)
class Lt(_BaseMetadata):
    """Lt(lt=x) implies that the value must be less than x.

    It can be used with any type that supports the ``<`` operator,
    including numbers, dates and times, strings, sets, and so on.
    """

    lt: SupportsLt

    def __get_pydantic_core_schema__(
        self, source_type, _handler
    ) -> core_schema.WrapValidatorFunctionSchema:
        string_schema = core_schema.any_schema()

        return core_schema.no_info_wrap_validator_function(
            cast(core_schema.NoInfoWrapValidatorFunction, self._validate), string_schema
        )

    def _validate(
        self, value: SupportsLt, handler: core_schema.ValidatorFunctionWrapHandler
    ) -> SupportsLt:
        if not value < self.lt:
            raise ValueError(f"Value {value} is not less than or equal to {self.lt}.")
        return value


@dataclass(frozen=True)
class Le(_BaseMetadata):
    """Le(le=x) implies that the value must be less than x.

    It can be used with any type that supports the ``<=`` operator,
    including numbers, dates and times, strings, sets, and so on.
    """

    le: SupportsLe

    def __get_pydantic_core_schema__(
        self, source_type, _handler
    ) -> core_schema.WrapValidatorFunctionSchema:
        string_schema = core_schema.any_schema()

        return core_schema.no_info_wrap_validator_function(
            cast(core_schema.NoInfoWrapValidatorFunction, self._validate), string_schema
        )

    def _validate(
        self, value: SupportsLe, handler: core_schema.ValidatorFunctionWrapHandler
    ) -> SupportsLe:
        if not value <= self.le:
            raise ValueError(f"Value {value} is not less than or equal to {self.le}.")
        return value


# @dataclass(frozen=True)
# class Re(_BaseMetadata):
#     """Re(pattern=s) implies that the value must be a string matching `s`.

#     It can be used with any string.
#     """

#     pattern: str | re.Pattern[str]


@dataclass(frozen=True)
class Re(_BaseMetadata):
    pattern: str | re.Pattern[str]

    def __post_init__(self) -> None:
        if isinstance(self.pattern, str):
            object.__setattr__(self, "pattern", re.compile(self.pattern))

    def __get_pydantic_core_schema__(
        self, source_type, _handler
    ) -> core_schema.WrapValidatorFunctionSchema:
        assert source_type is str, "Re may only be used on strings"

        string_schema = core_schema.str_schema()

        return core_schema.no_info_wrap_validator_function(
            cast(core_schema.NoInfoWrapValidatorFunction, self._validate), string_schema
        )

    def _validate(self, value: str, handler: core_schema.ValidatorFunctionWrapHandler) -> str:
        if not re.search(self.pattern, value):
            raise ValueError(f"String {value} does not match regex '{self.pattern!r}'.")
        return value


class SupportsIO(ABC):
    @classmethod
    @abstractmethod
    def read(cls, read_path: Path) -> Self: ...

    @abstractmethod
    def write(self, write_path: Path) -> None: ...


class SupportsJSONIO(ABC):
    @classmethod
    @abstractmethod
    def read_json_file(cls, read_path: Path) -> Self: ...

    @abstractmethod
    def write_json_file(self, write_path: Path) -> None: ...


def ensure_set(x: H | SequenceAlias[H]) -> set[H]:
    return set(cast(Iterable[H], x)) if isinstance(x, _SequenceAliasVar) else {x}


def ensure_tuple(x: T | SequenceAlias[T]) -> tuple[T, ...]:
    return tuple(cast(Iterable[T], x)) if isinstance(x, _SequenceAliasVar) else (x,)


def ensure_list(x: T | SequenceAlias[T]) -> list[T]:
    return list(cast(Iterable[T], x)) if isinstance(x, _SequenceAliasVar) else [x]


# ==============


class IdentityFunction[T](Protocol):
    def __call__(self, x: _T, /) -> _T: ...


# stable
class SupportsNext(Protocol[_T_co]):
    def __next__(self) -> _T_co: ...


class SupportsBool(Protocol):
    def __bool__(self) -> bool: ...


# Comparison protocols
class SupportsDunderLT(Protocol[_T_contra]):
    def __lt__(self, other: _T_contra, /) -> SupportsBool: ...


class SupportsDunderGT(Protocol[_T_contra]):
    def __gt__(self, other: _T_contra, /) -> SupportsBool: ...


class SupportsDunderLE(Protocol[_T_contra]):
    def __le__(self, other: _T_contra, /) -> SupportsBool: ...


class SupportsDunderGE(Protocol[_T_contra]):
    def __ge__(self, other: _T_contra, /) -> SupportsBool: ...


class SupportsAllComparisons(
    SupportsDunderLT[Any],
    SupportsDunderGT[Any],
    SupportsDunderLE[Any],
    SupportsDunderGE[Any],
    Protocol,
): ...


type SupportsRichComparison = SupportsDunderLT[Any] | SupportsDunderGT[Any]
SupportsRichComparisonT = TypeVar("SupportsRichComparisonT", bound=SupportsRichComparison)

C = TypeVar("C", bound=SupportsRichComparison)



import re
from collections.abc import Callable, Iterable

from .fp import endofilter
from .typing import areinstances


def as_pattern(s: str | re.Pattern[str]) -> re.Pattern[str]:
    return re.compile(s) if isinstance(s, str) else s


def re_bool_search(p: re.Pattern[str], s: str | None) -> bool:
    if not s:
        return False
    return bool(re.search(p, s))


def contains_re(pattern: re.Pattern[str], s: str | Iterable[str]) -> bool:
    if isinstance(s, str):
        return bool(re.search(pattern, s))
    for element in s:
        if re.search(pattern, element):
            return True
    return False


def re_group1(p: re.Pattern[str] | str, s: str | None) -> str | None:
    if s is None:
        return None
    if result := re.search(p, s):
        try:
            return result.group(1)
        except IndexError:
            return None
    return None


def re_groupdict(p: re.Pattern[str] | str, s: str | None) -> dict[str, str] | None:
    if s is None:
        return None
    if result := re.search(p, s):
        return result.groupdict()
    return None


def make_search(p: str | re.Pattern[str]) -> Callable[[str], bool]:
    if isinstance(p, str):
        p = re.compile(p)

    def search(s: str) -> bool:
        return bool(re.search(p, s))

    return search


def re_filter[T: list[str] | tuple[str, ...] | set[str]](p: str | re.Pattern[str], seq: T) -> T:
    return endofilter(make_search(p), seq)  # type: ignore


def re_split(expr: str | re.Pattern[str], s: str | None, maxsplit: int | None = None) -> list[str]:
    if s is None:
        return []
    segments = re.split(expr, s, maxsplit=maxsplit or 0)
    if not areinstances(segments, str):
        raise TypeError(
            "Function 're_split' may only return a list of strings."
        )  # pragma: no cover
    return segments


class Patterns:
    DATE: re.Pattern[str] = re.compile(r"^[12]\d\d\d-(0?[1-9]|1[012]|)-(0?[1-9]|[12]\d|3[01])$")
    DATE_STRICT: re.Pattern[str] = re.compile(
        r"^[12]\d\d\d-(0[1-9]|1[012]|)-(0[1-9]|[12]\d|3[01])$"
    )
    DATE_LOOSE: re.Pattern[str] = re.compile(r"(\d{2,4})[^\d](\d\d?)[^\d](\d\d?)")
    ID: re.Pattern[str] = re.compile(r"^[A-Za-z][A-Za-z0-9_]*$")
    IDS: re.Pattern[str] = re.compile(r"^([A-Za-z][A-Za-z0-9_]*)(,[A-Za-z][A-Za-z0-9_]*)*$")
    NATURAL: re.Pattern[str] = re.compile(r"^[1-9][0-9]*$")
    PROPORTION: re.Pattern[str] = re.compile(r"^0?\.[0-9]+$")
    PROPORTION_OR_2: re.Pattern[str] = re.compile(r"^0?\.[0-9]+$|^0?\.[0-9]+,0?\.[0-9]+$")



import operator
import re
from typing import Self


class Version:
    def __init__(self, major: int, minor: int, patch: int | str | None = None) -> None:
        self.major = major
        self.minor = minor
        self.patch = patch

    def __hash__(self) -> int:
        return hash((self.major, self.minor, self.patch))

    @property
    def patch_or_0(self) -> int | str:
        return self.patch or 0

    @classmethod
    def from_string(cls, version_string: str) -> Self | str:
        semver_pattern = re.compile(
            r".*?(?P<major>[0-9]+)\.(?P<minor>[0-9]+)(\.(?P<patch>.*))?"
            ""
        )
        search_result = re.search(semver_pattern, version_string)
        patch: str | int | None
        if search_result:
            gd = search_result.groupdict()
            _patch = gd.get("patch")
            try:
                patch = int(_patch or "")
            except ValueError:
                patch = _patch
            except TypeError:
                patch = _patch

            return cls(int(gd["major"]), int(gd["minor"]), patch)
        return version_string

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Version):
            raise TypeError
        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch_or_0 == other.patch_or_0
        )

    def minor_equal(self, other: "Version") -> bool:
        return self.major == other.major and self.minor == other.minor

    def meets_constraint(self, constraint: str) -> bool:
        pattern: re.Pattern[str] = re.compile(r"^ *(?P<op>\^|[\>\<\=]\=?) *(?P<version>.+) *")
        for segment in re.split(r" *, *", constraint):
            if not (search := re.search(pattern, segment)):
                raise ValueError(segment)
            gd = search.groupdict()
            if not isinstance(ver := Version.from_string(gd["version"]), Version):
                raise TypeError(gd["version"])
            op = {
                "==": operator.eq,
                "=": operator.eq,
                ">": operator.gt,
                "<": operator.lt,
                ">=": operator.ge,
                "<=": operator.le,
            }[gd["op"]]
            if not op(self, ver):
                return False

        return True

    def distance(self, other: "Version") -> str:
        def _distance(a: Version, b: Version) -> str:
            def replace_negative(i: int) -> str:
                return "?" if i < 0 else str(i)

            major = replace_negative(b.major - a.major)
            minor = replace_negative(b.minor - a.minor)
            try:
                if b.patch_or_0 == a.patch_or_0:
                    _patch = 0
                else:
                    _patch = b.patch_or_0 - a.patch_or_0  # type: ignore
                patch = replace_negative(_patch)
            except TypeError:
                patch = "?"
            return f"{major}.{minor}.{patch}"

        if self > other:
            return _distance(other, self)
        elif other > self:
            return _distance(self, other)
        return "?.?.?"

    def __sub__(self: Self, other: "Version") -> "Version":
        try:
            patch_diff = self.patch_or_0 - other.patch_or_0  # type: ignore
        except TypeError:
            patch_diff = "?"
        return self.__class__(
            self.major - other.major,
            self.minor - other.minor,
            patch_diff,
        )

    def __ge__(self, other: "Version") -> bool:
        if (self.major, self.minor, self.patch_or_0) < (other.major, other.minor, self.patch_or_0):
            return False
        if self.minor_equal(other) and (type(self.patch_or_0) != type(self.patch_or_0)):  # noqa: E721
            return False
        return (self.major, self.minor, self.patch_or_0) >= (
            other.major,
            other.minor,
            other.patch_or_0,
        )

    def __gt__(self, other: "Version") -> bool:
        if (self.major, self.minor, str(self.patch_or_0)) <= (
            other.major,
            other.minor,
            str(other.patch_or_0),
        ):
            return False
        if self.minor_equal(other):
            if type(self.patch_or_0) != type(other.patch_or_0):  # noqa: E721
                return False

        return (self.major, self.minor, self.patch_or_0) > (
            other.major,
            other.minor,
            other.patch_or_0,
        )



from . import (
    collections,
    color,
    comparison,
    converters,
    depsort,
    display,
    ds,
    enums,
    exceptions,
    fp,
    imports,
    io,
    markers,
    merge,
    modtime,
    num,
    path,
    perf,
    pydantic,
    re,
    semver,
    shell,
    string,
    timestamps,
    tuples,
    types,
    typing,
)

__all__ = (
    "collections",
    "color",
    "comparison",
    "converters",
    "depsort",
    "display",
    "ds",
    "enums",
    "exceptions",
    "fp",
    "imports",
    "io",
    "markers",
    "merge",
    "modtime",
    "num",
    "path",
    "perf",
    "pydantic",
    "re",
    "semver",
    "shell",
    "string",
    "timestamps",
    "tuples",
    "types",
    "typing",
)



from datetime import datetime
from pathlib import Path


def make_timestamp(sep: str = "T", places: int = 0) -> str:
    dt = datetime.now()
    if places > 0:
        idx: int | None = min(0, int(places) - 6) or None
        return dt.strftime(f"%Y-%m-%d{sep}%H:%M:%S.%f")[:idx]
    return dt.strftime(f"%Y-%m-%d{sep}%H:%M:%S")


def insert_timestamp(p: Path | str, sep: str = "T", places: int = 0) -> Path:
    new_path = str(p)

    if "." in Path(p).name:
        base, suffix = new_path.rsplit(".", 1)
        return Path(f"{base}__{make_timestamp(sep=sep, places=places)}.{suffix}")

    return Path(f"{new_path}__{make_timestamp(sep=sep, places=places)}")



from collections.abc import Callable, Hashable, Iterable, Sequence
from functools import reduce
from typing import TypeVar, overload

__all__ = (
    "dfilter",
    "dmap",
    "endofilter",
    "endomap",
    "fold_dictionaries",
    "identity",
    "kfilter",
    "kmap",
    "kvfilter",
    "kvmap",
    "lfilter",
    "lmap",
    # "negate",
    "sfilter",
    "smap",
    "split_head",
    "tail",
    "tfilter",
    "tmap",
    "vfilter",
    "vmap",
)

T = TypeVar("T")
Tup = TypeVar("Tup", bound=tuple[object, ...])
Seq = TypeVar("Seq", bound=list | tuple)
KPre = TypeVar("KPre")
VPre = TypeVar("VPre")
KPost = TypeVar("KPost")
VPost = TypeVar("VPost")
TPost = TypeVar("TPost")
TPre = TypeVar("TPre")
K = TypeVar("K", bound=Hashable)
V = TypeVar("V")
type Filterer[T] = Callable[[T], bool]
type DictLike[K, V] = dict[K, V] | Iterable[tuple[K, V]]


def negate[T: object](f: Callable[[T], object]) -> Callable[[T], bool]:
    def inner(ob: T) -> bool:
        return not f(ob)
    
    return inner


@overload
def endomap(
    callable_: Callable[[TPre], TPost], sequence: list[TPre]
) -> list[TPost]: ...  # pragma: no cover
@overload
def endomap(
    callable_: Callable[[TPre], TPost], sequence: set[TPre]
) -> set[TPost]: ...  # pragma: no cover
@overload
def endomap(
    callable_: Callable[[TPre], TPost], sequence: tuple[TPre, ...]
) -> tuple[TPost, ...]: ...  # pragma: no cover
def endomap(callable_, sequence):
    return type(sequence)(map(callable_, sequence))


@overload
def endofilter(
    callable_: Callable[[T], bool], sequence: list[T]
) -> list[T]: ...  # pragma: no cover
@overload
def endofilter(callable_: Callable[[T], bool], sequence: set[T]) -> set[T]: ...  # pragma: no cover
@overload
def endofilter(
    callable_: Callable[[T], bool], sequence: tuple[T, ...]
) -> tuple[T, ...]: ...  # pragma: no cover
def endofilter(callable_, sequence):
    return type(sequence)(filter(callable_, sequence))


def _handle_dictlike[K, V](d: DictLike[K, V]) -> Iterable[tuple[K, V]]:
    if isinstance(d, dict):
        tmp = tuple(d.items())
        return tmp  # type: ignore
    return d


def lmap(callable_: Callable[[TPre], TPost], iterable: Iterable[TPre]) -> list[TPost]:
    return list(map(callable_, iterable))


def smap(callable_: Callable[[TPre], TPost], iterable: Iterable[TPre]) -> set[TPost]:
    return set(map(callable_, iterable))


def tmap(callable_: Callable[[TPre], TPost], iterable: Iterable[TPre]) -> tuple[TPost, ...]:
    return tuple(map(callable_, iterable))


def vmap(callable_: Callable[[TPre], TPost], dictionary: DictLike[K, TPre]) -> dict[K, TPost]:
    return {k: callable_(v) for k, v in _handle_dictlike(dictionary)}


def kmap(callable_: Callable[[TPre], TPost], dictionary: DictLike[TPre, V]) -> dict[TPost, V]:
    return {callable_(k): v for k, v in _handle_dictlike(dictionary)}


def dmap(
    callable_: Callable[[TPre], TPost], dictionary: DictLike[TPre, TPre]
) -> dict[TPost, TPost]:
    return {callable_(k): callable_(v) for k, v in _handle_dictlike(dictionary)}


def kvmap(
    callable_: Callable[[tuple[KPre, VPre]], tuple[KPost, VPost]], d: DictLike[KPre, VPre]
) -> dict[KPost, VPost]:
    return dict(map(callable_, _handle_dictlike(d)))


def lfilter(filterer: Filterer[T], iterable: Iterable[T]) -> list[T]:
    return list(filter(filterer, iterable))


def sfilter(filterer: Filterer[T], iterable: Iterable[T]) -> set[T]:
    return set(filter(filterer, iterable))


def tfilter(filterer: Filterer[T], iterable: Iterable[T]) -> tuple[T, ...]:
    return tuple(filter(filterer, iterable))


def vfilter(filterer: Filterer[V], dictionary: dict[K, V]) -> dict[K, V]:
    return {k: v for k, v in dictionary.items() if filterer(v)}


def kfilter(filterer: Filterer[K], dictionary: dict[K, V]) -> dict[K, V]:
    return {k: v for k, v in dictionary.items() if filterer(k)}


def dfilter(filterer: Filterer[T], dictionary: DictLike[T, T]) -> dict[T, T]:
    return {k: v for k, v in _handle_dictlike(dictionary) if filterer(k) and filterer(v)}


def kvfilter(callable_: Callable[[tuple[K, V]], bool], d: DictLike[K, V]) -> dict[K, V]:
    return dict(filter(callable_, _handle_dictlike(d)))


def identity[T](x: T) -> T:
    return x


def fold_dictionaries[K, V](dicts: Iterable[dict[K, V]]) -> dict[K, V]:
    def _or(dict1: dict[K, V], dict2: dict[K, V]) -> dict[K, V]:
        return dict1 | dict2

    return reduce(_or, dicts)


def split_head[T](seq: Sequence[T]) -> tuple[T, Sequence[T]]:
    if not seq:
        raise ValueError(f"List {seq} is invalid because it cannot be properly split.")
    return seq[0], seq[1:]


def tail[T](seq: Sequence[T]) -> Sequence[T]:
    return seq[1:]
